from unittest import TestCase

from tests.fixtures import clear_dir, init_dir

from ...repositories.project.project import (
    ConditionStage,
    JobStage,
    ProjectRepository,
    ScriptStage,
    WorkflowTransition,
)
from ...repositories.stage_run import LocalStageRunRepository
from .kanban import (
    ColumnStage,
    DataRequest,
    KanbanColumn,
    KanbanController,
    KanbanData,
)


class KanbanTests(TestCase):
    def setUp(self):
        self.root = init_dir()

        self.stage_run_repository = LocalStageRunRepository()
        self.project_repository = ProjectRepository
        self.controller = KanbanController(
            stage_run_repository=self.stage_run_repository,
            project_repository=self.project_repository,
        )

    def tearDown(self) -> None:
        clear_dir(self.root)

    def test_get_data_empty(self):
        data = self.controller.get_data(
            DataRequest.from_dict(
                {
                    "selection": [],
                    "filter": {},
                }
            )
        )

        self.assertEqual(
            data, KanbanData(columns=[], next_stage_options=[], not_found_stages=[])
        )

    def test_get_data_single_stage_no_selection(self):
        project = ProjectRepository.load()

        job = JobStage(
            id="job",
            title="Job",
            file="job.py",
            schedule="* * * * *",
            workflow_position=(0, 0),
            workflow_transitions=[],
        )
        project.add_stage(job)
        ProjectRepository.save(project)

        data = self.controller.get_data(
            DataRequest.from_dict(
                {
                    "selection": [],
                    "filter": {},
                }
            )
        )

        self.assertEqual(
            data,
            KanbanData(
                columns=[],
                next_stage_options=[
                    ColumnStage(
                        id="job",
                        title="Job",
                        path=None,
                        can_be_started=True,
                        type="job",
                    )
                ],
                not_found_stages=[],
            ),
        )

    def test_get_data_single_stage_with_selection(self):
        self.maxDiff = None
        project = ProjectRepository.load()

        job = JobStage(
            id="job",
            title="Job",
            file="job.py",
            schedule="* * * * *",
            workflow_position=(0, 0),
            workflow_transitions=[],
        )

        project.add_stage(job)
        ProjectRepository.save(project)

        data = self.controller.get_data(
            DataRequest.from_dict(
                {
                    "selection": [
                        {
                            "stage_id": "job",
                            "limit": 10,
                            "offset": 0,
                        }
                    ],
                    "filter": {},
                }
            )
        )

        self.assertEqual(
            data,
            KanbanData(
                columns=[
                    KanbanColumn(
                        selected_stage=ColumnStage.create(job),
                        stages=[ColumnStage.create(job)],
                        stage_run_cards=[],
                        total_count=0,
                    )
                ],
                next_stage_options=[],
                not_found_stages=[],
            ),
        )

    def test_get_data_multiple_stages_no_selection(self):
        self.maxDiff = None
        project = ProjectRepository.load()

        script = ScriptStage(
            id="script",
            file="script.py",
            title="Script",
            workflow_position=(0, 0),
            workflow_transitions=[],
        )
        condition = ConditionStage(
            id="condition",
            variable_name="title",
            type_name="condition",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="transition",
                    target_id="script",
                    target_type="script",
                    type="conditions:patternMatched",
                    condition_value="true",
                )
            ],
        )

        job = JobStage(
            id="job",
            title="Job",
            file="job.py",
            schedule="* * * * *",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="transition",
                    target_id="condition",
                    target_type="condition",
                    type="jobs:finished",
                    condition_value=None,
                )
            ],
        )

        project.add_stage(job)
        project.add_stage(condition)
        project.add_stage(script)
        ProjectRepository.save(project)

        data = self.controller.get_data(
            DataRequest.from_dict(
                {
                    "selection": [],
                    "filter": {},
                }
            )
        )

        self.assertEqual(
            data,
            KanbanData(
                columns=[],
                next_stage_options=[
                    ColumnStage.create(job),
                    ColumnStage.create(script),
                    ColumnStage.create(condition),
                ],
                not_found_stages=[],
            ),
        )

    def test_get_data_multiple_stages_single_selection(self):
        self.maxDiff = None
        project = ProjectRepository.load()

        script = ScriptStage(
            id="script",
            file="script.py",
            title="Script",
            workflow_position=(0, 0),
            workflow_transitions=[],
        )
        condition = ConditionStage(
            id="condition",
            variable_name="title",
            type_name="condition",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="transition",
                    target_id="script",
                    target_type="script",
                    type="conditions:patternMatched",
                    condition_value="true",
                )
            ],
        )

        job = JobStage(
            id="job",
            title="Job",
            file="job.py",
            schedule="* * * * *",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="transition",
                    target_id="condition",
                    target_type="condition",
                    type="jobs:finished",
                    condition_value=None,
                )
            ],
        )

        project.add_stage(job)
        project.add_stage(condition)
        project.add_stage(script)
        ProjectRepository.save(project)

        data = self.controller.get_data(
            DataRequest.from_dict(
                {
                    "selection": [
                        {
                            "stage_id": "job",
                            "limit": 10,
                            "offset": 0,
                        }
                    ],
                    "filter": {},
                }
            )
        )

        self.assertEqual(
            data,
            KanbanData(
                columns=[
                    KanbanColumn(
                        selected_stage=ColumnStage.create(job),
                        stages=[
                            ColumnStage.create(script),
                            ColumnStage.create(condition),
                            ColumnStage.create(job),
                        ],
                        stage_run_cards=[],
                        total_count=0,
                    )
                ],
                next_stage_options=[
                    ColumnStage.create(script),
                    ColumnStage.create(condition),
                ],
                not_found_stages=[],
            ),
        )

    def test_get_data_multiple_stages_multiple_selection(self):
        self.maxDiff = None
        project = ProjectRepository.load()

        script = ScriptStage(
            id="script",
            file="script.py",
            title="Script",
            workflow_position=(0, 0),
            workflow_transitions=[],
        )
        condition = ConditionStage(
            id="condition",
            variable_name="title",
            type_name="condition",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="transition",
                    target_id="script",
                    target_type="script",
                    type="conditions:patternMatched",
                    condition_value="true",
                )
            ],
        )

        job = JobStage(
            id="job",
            title="Job",
            file="job.py",
            schedule="* * * * *",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="transition",
                    target_id="condition",
                    target_type="condition",
                    type="jobs:finished",
                    condition_value=None,
                )
            ],
        )

        project.add_stage(job)
        project.add_stage(condition)
        project.add_stage(script)
        ProjectRepository.save(project)

        data = self.controller.get_data(
            DataRequest.from_dict(
                {
                    "selection": [
                        {
                            "stage_id": "job",
                            "limit": 10,
                            "offset": 0,
                        },
                        {
                            "stage_id": "condition",
                            "limit": 10,
                            "offset": 0,
                        },
                    ],
                    "filter": {},
                }
            )
        )

        self.assertEqual(
            data,
            KanbanData(
                columns=[
                    KanbanColumn(
                        selected_stage=ColumnStage.create(job),
                        stages=[
                            ColumnStage.create(script),
                            ColumnStage.create(job),
                        ],
                        stage_run_cards=[],
                        total_count=0,
                    ),
                    KanbanColumn(
                        selected_stage=ColumnStage.create(condition),
                        stages=[
                            ColumnStage.create(script),
                            ColumnStage.create(condition),
                        ],
                        stage_run_cards=[],
                        total_count=0,
                    ),
                ],
                next_stage_options=[ColumnStage.create(script)],
                not_found_stages=[],
            ),
        )

    def test_get_data_multiple_multiple_initial_with_selection(self):
        self.maxDiff = None
        project = ProjectRepository.load()

        script = ScriptStage(
            id="script",
            file="script.py",
            title="Script",
            workflow_position=(0, 0),
            workflow_transitions=[],
        )
        condition = ConditionStage(
            id="condition",
            variable_name="title",
            type_name="condition",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="transition",
                    target_id="script",
                    target_type="script",
                    type="conditions:patternMatched",
                    condition_value="true",
                )
            ],
        )

        job1 = JobStage(
            id="job1",
            title="Job1",
            file="job1.py",
            schedule="* * * * *",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="transition",
                    target_id="condition",
                    target_type="condition",
                    type="jobs:finished",
                    condition_value=None,
                )
            ],
        )

        job2 = JobStage(
            id="job2",
            title="Job2",
            file="job2.py",
            schedule="* * * * *",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="transition",
                    target_id="condition",
                    target_type="condition",
                    type="jobs:finished",
                    condition_value=None,
                )
            ],
        )

        project.add_stage(job1)
        project.add_stage(job2)
        project.add_stage(condition)
        project.add_stage(script)
        ProjectRepository.save(project)

        data = self.controller.get_data(
            DataRequest.from_dict(
                {
                    "selection": [],
                    "filter": {},
                }
            )
        )

        self.assertEqual(
            data,
            KanbanData(
                columns=[],
                next_stage_options=[
                    ColumnStage.create(job1),
                    ColumnStage.create(job2),
                    ColumnStage.create(script),
                    ColumnStage.create(condition),
                ],
                not_found_stages=[],
            ),
        )

    def test_get_data_multiple_multiple_initial_with_multiple_selection(self):
        self.maxDiff = None
        project = ProjectRepository.load()

        script = ScriptStage(
            id="script",
            file="script.py",
            title="Script",
            workflow_position=(0, 0),
            workflow_transitions=[],
        )
        condition = ConditionStage(
            id="condition",
            variable_name="title",
            type_name="condition",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="transition",
                    target_id="script",
                    target_type="script",
                    type="conditions:patternMatched",
                    condition_value="true",
                )
            ],
        )

        job1 = JobStage(
            id="job1",
            title="Job1",
            file="job1.py",
            schedule="* * * * *",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="transition",
                    target_id="condition",
                    target_type="condition",
                    type="jobs:finished",
                    condition_value=None,
                )
            ],
        )

        job2 = JobStage(
            id="job2",
            title="Job2",
            file="job2.py",
            schedule="* * * * *",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="transition",
                    target_id="condition",
                    target_type="condition",
                    type="jobs:finished",
                    condition_value=None,
                )
            ],
        )

        project.add_stage(job1)
        project.add_stage(job2)
        project.add_stage(condition)
        project.add_stage(script)
        ProjectRepository.save(project)

        data = self.controller.get_data(
            DataRequest.from_dict(
                {
                    "selection": [
                        {
                            "stage_id": "job1",
                            "limit": 10,
                            "offset": 0,
                        },
                        {
                            "stage_id": "condition",
                            "limit": 10,
                            "offset": 0,
                        },
                    ],
                }
            )
        )

        self.assertEqual(
            data,
            KanbanData(
                columns=[
                    KanbanColumn(
                        selected_stage=ColumnStage.create(job1),
                        stages=[
                            ColumnStage.create(job2),
                            ColumnStage.create(script),
                            ColumnStage.create(job1),
                        ],
                        stage_run_cards=[],
                        total_count=0,
                    ),
                    KanbanColumn(
                        selected_stage=ColumnStage.create(condition),
                        stages=[
                            ColumnStage.create(job2),
                            ColumnStage.create(script),
                            ColumnStage.create(condition),
                        ],
                        stage_run_cards=[],
                        total_count=0,
                    ),
                ],
                next_stage_options=[
                    ColumnStage.create(job2),
                    ColumnStage.create(script),
                ],
                not_found_stages=[],
            ),
        )

    def test_return_not_found_stages(self):
        self.maxDiff = None
        project = ProjectRepository.load()

        ProjectRepository.save(project)

        data = self.controller.get_data(
            DataRequest.from_dict(
                {
                    "selection": [
                        {
                            "stage_id": "script",
                            "limit": 10,
                            "offset": 0,
                        }
                    ],
                    "filter": {},
                }
            )
        )

        self.assertEqual(
            data,
            KanbanData(
                columns=[],
                next_stage_options=[],
                not_found_stages=["script"],
            ),
        )

    def test_get_job(self):
        project = ProjectRepository.load()

        job = JobStage(
            id="job",
            title="Job",
            file="job.py",
            schedule="* * * * *",
            workflow_position=(0, 0),
            workflow_transitions=[],
        )
        project.add_stage(job)
        ProjectRepository.save(project)

        rtvd = self.controller.get_job("job")

        self.assertEqual(rtvd, job)



from dataclasses import dataclass
from datetime import date, datetime
from unittest import TestCase
from uuid import UUID

from ..interface.sdk.forms.page_response import PageResponse
from ..interface.sdk.forms.step import StepsResponse
from ..widgets.response_types import FileResponse, PhoneResponse
from .json import to_json_serializable


class TestToJsonSerializable(TestCase):
    def test_int(self):
        self.assertEqual(to_json_serializable(1), 1)

    def test_float(self):
        self.assertEqual(to_json_serializable(1.0), 1.0)

    def test_none(self):
        self.assertEqual(to_json_serializable(None), None)

    def test_string(self):
        self.assertEqual(to_json_serializable("test"), "test")

    def test_list(self):
        self.assertEqual(to_json_serializable([1, 2, 3]), [1, 2, 3])

    def test_tuple(self):
        self.assertEqual(to_json_serializable((1, 2, 3)), [1, 2, 3])

    def test_set(self):
        self.assertEqual(to_json_serializable({1, 2, 3}), [1, 2, 3])

    def test_dict(self):
        self.assertEqual(to_json_serializable({"a": 1, "b": 2}), {"a": 1, "b": 2})

    def test_page_response(self):
        self.assertEqual(
            to_json_serializable(PageResponse(data={"foo": "bar"}, action="baz")),
            {"foo": "bar"},
        )

    def test_page_response_complex(self):
        self.assertEqual(
            to_json_serializable(
                PageResponse(
                    data={
                        "datetime": datetime(2021, 1, 1, 12, 0, 0),
                        "uuid": UUID("123e4567-e89b-12d3-a456-426614174000"),
                    },
                    action="foo",
                )
            ),
            {
                "datetime": "2021-01-01T12:00:00",
                "uuid": "123e4567-e89b-12d3-a456-426614174000",
            },
        )

    def test_steps_response(self):
        steps_response = StepsResponse()
        steps_response.append(PageResponse(data={"a": 1, "b": 2}, action="baz"))
        steps_response.append(PageResponse(data={"b": 3, "c": 4}, action="foo"))
        self.assertEqual(to_json_serializable(steps_response), {"a": 1, "b": 3, "c": 4})

    def test_dataclass(self):
        @dataclass
        class Foo:
            a: int
            b: str

        self.assertEqual(to_json_serializable(Foo(1, "bar")), {"a": 1, "b": "bar"})

    def test_datetime(self):
        now = datetime(2021, 1, 1, 12, 0, 0)
        self.assertEqual(to_json_serializable(now), "2021-01-01T12:00:00")

    def test_date(self):
        now = date(2021, 1, 1)
        self.assertEqual(to_json_serializable(now), "2021-01-01")

    def test_uuid(self):
        uuid = UUID("123e4567-e89b-12d3-a456-426614174000")
        self.assertEqual(
            to_json_serializable(uuid), "123e4567-e89b-12d3-a456-426614174000"
        )

    def test_file_response(self):
        file_response = FileResponse("foo.txt")
        self.assertEqual(to_json_serializable(file_response), "foo.txt")

    def test_phone_response(self):
        phone_response = PhoneResponse(
            country_code="55",
            masked="+55 (21) 9999-9999",
            national_number="21999999999",
            raw="5521999999999",
        )
        self.assertEqual(to_json_serializable(phone_response), "+55 (21) 9999-9999")



from typing import Literal

import pkg_resources
import sentry_sdk

from .utils import is_testing


class AbstraLogger:
    @staticmethod
    def init(environment: Literal["cloud", "local"]):
        if is_testing():
            return

        try:
            sentry_sdk.init(
                dsn="https://9bbccd1a46ddb8a563483c6afc61ca35@o1317386.ingest.us.sentry.io/4507024713383936",
                traces_sample_rate=1.0,
                profiles_sample_rate=1.0,
                environment=environment,
                enable_tracing=True,
                release=pkg_resources.get_distribution("abstra").version,
            )
        except Exception:
            print("Error reporting has been turned off.")

    @staticmethod
    def capture_exception(exception: Exception):
        if is_testing():
            return

        sentry_sdk.capture_exception(exception)

    @staticmethod
    def capture_message(message: str):
        if is_testing():
            return

        sentry_sdk.capture_message(message)



import webbrowser
from pathlib import Path
from typing import List, Set

from ...contracts_generated import AbstraLibApiEditorEnvVarsModel
from ...repositories.env_vars import EnvVarsRepository
from ..linter import LinterFix, LinterIssue, LinterRule


class AddEnvToEnvFile(LinterFix):
    def __init__(self, env_var: str):
        self.label = f"Add {env_var} to .env file"
        self.env_var = env_var

    def fix(self):
        env_var = self.env_var
        EnvVarsRepository.set(
            AbstraLibApiEditorEnvVarsModel(key=env_var, value=">> REPLACE ME <<")
        )
        env_uri = EnvVarsRepository.get_env_var_path().absolute().as_uri()
        webbrowser.open(env_uri)


class EnvInCodeNotInEnvFile(LinterIssue):
    def __init__(self, filename: Path, lineno: int, env_var: str):
        self.label = f"Env var {env_var} is used in {filename}:{lineno} but not defined in .env file"
        self.fixes = [AddEnvToEnvFile(env_var)]


class MissingEnv(LinterRule):
    label: str = "Missing env vars"
    type: str = "info"

    def find_issues(self) -> List[LinterIssue]:
        env_vars_in_code_dict = EnvVarsRepository.get_env_vars_in_code()
        env_vars_in_code: Set[str] = set(env_vars_in_code_dict.keys())
        env_vars_in_env_file: Set[str] = set(
            [ev.key for ev in EnvVarsRepository.list()]
        )
        missing_env_vars = env_vars_in_code - env_vars_in_env_file

        issues = []

        for missing_env_var in missing_env_vars:
            code_paths: Set[str] = set()
            for filename, expr in env_vars_in_code_dict[missing_env_var]:
                code_path = f"{filename}:{expr.lineno}"
                if code_path not in code_paths:
                    code_paths.add(code_path)
                    issues.append(
                        EnvInCodeNotInEnvFile(filename, expr.lineno, missing_env_var)
                    )
        return issues



from unittest import TestCase

from tests.fixtures import clear_dir, init_dir

from ...repositories.project.project import (
    ProjectRepository,
    VisualizationItem,
    VisualizationSettings,
)
from .visualizations import VisualizationsController


class VisualizationsTest(TestCase):
    def setUp(self):
        self.controller = VisualizationsController()
        self.root = init_dir()

    def tearDown(self) -> None:
        clear_dir(self.root)

    def test_loads_correctly_from_empty_json(self):
        project = ProjectRepository.load()
        self.assertEqual(project.visualization, VisualizationSettings.create())

    def test_loads_correctly_from_json_content(self):
        visualizations = [{"name": "foo", "type": "text"}]
        VisualizationsController().put_visualizations(visualizations)

        project = ProjectRepository.load()
        self.assertEqual(
            project.visualization.items[0], VisualizationItem(name="foo", type="text")
        )

    def test_generates_correct_dict(self):
        ProjectRepository.load()

        self.controller.put_visualizations([{"name": "foo", "type": "text"}])

        received = self.controller.get_visualizations()
        self.assertEqual(
            received,
            [{"name": "foo", "type": "text"}],
        )



import copy
import json
import uuid
from abc import ABC, abstractmethod
from datetime import datetime
from typing import Any, Dict, List, Mapping, Optional, Union

import requests
from pydantic import ConfigDict, Field
from pydantic.dataclasses import dataclass

from ..utils import serialize
from ..utils.datetime import from_utc_iso_string
from ..utils.environment import SIDECAR_HEADERS, SIDECAR_URL
from ..utils.validate import validate_json

end_status = ["failed", "finished", "abandoned"]

status_transitions = {
    "waiting": ["running"],
    "running": end_status,
}


@dataclass
class Pagination:
    limit: int
    offset: int

    @staticmethod
    def from_dict(data: dict) -> "Pagination":
        return Pagination(
            limit=data.get("limit", 10),
            offset=data.get("offset", 0),
        )

    def to_dict(self) -> dict:
        return dict((k, v) for k, v in self.__dict__.items() if v is not None)


@dataclass
class PaginatedListResponse:
    stage_runs: List["StageRun"]
    total_count: int

    @staticmethod
    def from_dict(data: dict) -> "PaginatedListResponse":
        return PaginatedListResponse(
            stage_runs=[StageRun.from_dict(dto) for dto in data.get("stageRuns", [])],
            total_count=data.get("totalCount", 0),
        )

    def to_dict(self) -> dict:
        return dict((k, v) for k, v in self.__dict__.items() if v is not None)


@dataclass
class GetStageRunByQueryFilter:
    stage: Optional[Union[str, List[str]]] = None
    assignee: Optional[str] = None
    parent_id: Optional[str] = None
    status: Optional[str] = None
    data: Optional[Dict] = None

    @staticmethod
    def from_dict(data: dict) -> "GetStageRunByQueryFilter":
        return GetStageRunByQueryFilter(
            stage=data.get("stage", None),
            assignee=data.get("assignee", None),
            parent_id=data.get("parent_id", None),
            status=data.get("status", None),
            data=data.get("data", None),
        )

    def to_dict(self) -> dict:
        return dict((k, v) for k, v in self.__dict__.items() if v is not None)


@dataclass(config=ConfigDict(populate_by_name=True))
class StageRun:
    id: str
    stage: str
    data: dict
    status: str
    created_at: datetime = Field(serialization_alias="createdAt")
    updated_at: datetime = Field(default=None, serialization_alias="updatedAt")
    parent_id: Optional[str] = Field(default=None, serialization_alias="parentId")
    execution_id: Optional[str] = Field(default=None, serialization_alias="executionId")

    @staticmethod
    def from_dict(data: dict) -> "StageRun":
        return StageRun(
            id=data["id"],
            stage=data["stage"],
            data=data["data"],
            status=data["status"],
            parent_id=data.get("parentId"),
            execution_id=data.get("executionId"),
            created_at=from_utc_iso_string(data["createdAt"]),
            updated_at=from_utc_iso_string(data["updatedAt"]),
        )

    def __getitem__(self, key):
        return self.data.get(key)

    def __setitem__(self, key, value):
        self.data[key] = value

    def get(self, key, default=None):
        return self.data.get(key, default)

    @property
    def is_end(self):
        return self.status in end_status

    @property
    def is_abandoned(self):
        return self.status == "abandoned"

    def to_dto(self):
        return {
            "id": self.id,
            "stage": self.stage,
            "data": self.data,
            "status": self.status,
            "parentId": self.parent_id,
            "executionId": self.execution_id,
            "createdAt": self.created_at.isoformat(),
            "updatedAt": self.updated_at.isoformat(),
        }

    def clone(self):
        return StageRun(**copy.deepcopy(self.__dict__))

    def clone_to_waiting(self):
        new_stage_run = self.clone()
        new_stage_run.id = str(uuid.uuid4())
        new_stage_run.status = "waiting"
        new_stage_run.created_at = datetime.now()
        new_stage_run.updated_at = datetime.now()
        return new_stage_run


class StageRunRepository(ABC):
    valid_filter_keys = ["data", "stage", "status", "parent_id"]

    def validate_filter_keys(self, filter: Dict) -> None:
        for key in filter.keys():
            if key not in self.valid_filter_keys:
                raise Exception(f"Invalid filter key {key}")

    def hydrate_next_dto(self, parent: StageRun, dtos: List[Dict]) -> List[Dict]:
        next_dtos = []

        for dto in dtos:
            stage = dto.get("stage")
            data = {**parent.data, **dto.get("data", {})}
            next_dtos.append(dict(stage=stage, data=data))

        return next_dtos

    @abstractmethod
    def find(self, filter: GetStageRunByQueryFilter) -> List[StageRun]:
        raise NotImplementedError()

    @abstractmethod
    def find_ancestors(self, id: str) -> List[StageRun]:
        raise NotImplementedError()

    @abstractmethod
    def find_leaves(
        self, filter: GetStageRunByQueryFilter, pagination: Pagination
    ) -> PaginatedListResponse:
        raise NotImplementedError()

    @abstractmethod
    def find_past(
        self, filter: GetStageRunByQueryFilter, pagination: Pagination
    ) -> PaginatedListResponse:
        raise NotImplementedError()

    @abstractmethod
    def get(self, id: str) -> StageRun:
        raise NotImplementedError()

    @abstractmethod
    def change_state(
        self, id: str, status: str, execution_id: Optional[str] = None
    ) -> bool:
        raise NotImplementedError()

    @abstractmethod
    def create_initial(self, stage: str, data: dict = {}) -> StageRun:
        raise NotImplementedError()

    @abstractmethod
    def create_next(self, parent: StageRun, dtos: List[Dict]) -> List[StageRun]:
        raise NotImplementedError()

    @abstractmethod
    def fork(self, stage_run: StageRun, custom_thread_data: Optional[str]) -> StageRun:
        raise NotImplementedError()

    @abstractmethod
    def update_data(self, stage_run_id: str, data: dict) -> bool:
        raise NotImplementedError()


class LocalStageRunRepository(StageRunRepository):
    _stage_runs: List[StageRun]

    def __init__(self):
        self._stage_runs = []
        self._locks = []

    def clear(self):
        self._stage_runs = []
        self._locks = []

    def get(self, id: str) -> StageRun:
        for stage_run in self._stage_runs:
            if stage_run.id == id:
                return stage_run

        raise Exception(f"StageRun with id {id} not found")

    def find(self, filter: GetStageRunByQueryFilter) -> List[StageRun]:
        results: List[StageRun] = []

        for stage_run in self._stage_runs:
            if filter.stage:
                if isinstance(filter.stage, list):
                    if stage_run.stage not in filter.stage:
                        continue
                else:
                    if stage_run.stage != filter.stage:
                        continue

            if filter.data and not self._compare_data(stage_run.data, filter.data):
                continue

            if filter.status and stage_run.status != filter.status:
                continue

            if filter.parent_id and stage_run.parent_id != filter.parent_id:
                continue

            results.append(stage_run)

        return results

    def _compare_data(self, data: dict, filter: dict) -> bool:
        filtered_data = {k: v for k, v in data.items() if k in filter}
        return serialize(filtered_data, sort_keys=True) == serialize(
            filter, sort_keys=True
        )

    def find_ancestors(self, id: str) -> List[StageRun]:
        ancestors = []
        stage_run = self.get(id)
        ancestors.append(stage_run)
        while stage_run.parent_id is not None:
            stage_run = self.get(stage_run.parent_id)
            ancestors.insert(0, stage_run)
        return ancestors

    def find_leaves(
        self, filter: GetStageRunByQueryFilter, pagination: Pagination
    ) -> PaginatedListResponse:
        filter_matches = self.find(filter)
        parent_ids = set(u.parent_id for u in self._stage_runs)

        leaves = sorted(
            [
                stage_run
                for stage_run in filter_matches
                if stage_run.id not in parent_ids
            ],
            key=lambda x: x.created_at,
            reverse=True,
        )

        total_count = len(leaves)
        if pagination.offset >= total_count:
            return PaginatedListResponse(stage_runs=[], total_count=total_count)

        return PaginatedListResponse(
            stage_runs=leaves[pagination.offset : pagination.limit],
            total_count=total_count,
        )

    def find_past(
        self, filter: GetStageRunByQueryFilter, pagination: Pagination
    ) -> PaginatedListResponse:
        filter_matches = self.find(filter)
        parent_ids = set(u.parent_id for u in self._stage_runs)

        past = sorted(
            [stage_run for stage_run in filter_matches if stage_run.id in parent_ids],
            key=lambda x: x.created_at,
            reverse=True,
        )

        total_count = len(past)
        if pagination.offset >= total_count:
            return PaginatedListResponse(stage_runs=[], total_count=total_count)

        return PaginatedListResponse(
            stage_runs=past[pagination.offset : pagination.limit],
            total_count=total_count,
        )

    def create_initial(self, stage: str, data: dict = {}) -> StageRun:
        stage_run = StageRun(
            id=str(uuid.uuid4()),
            status="waiting",
            stage=stage,
            data=data,
            parent_id=None,
            created_at=datetime.now(),
            updated_at=datetime.now(),
        )

        self._stage_runs.append(stage_run)

        return stage_run

    def create_next(self, parent: StageRun, dtos: List[Dict]) -> List[StageRun]:
        created = []
        for dto in self.hydrate_next_dto(parent, dtos):
            stage_run = StageRun(
                id=str(uuid.uuid4()),
                parent_id=parent.id,
                status="waiting",
                **dto,
                created_at=datetime.now(),
                updated_at=datetime.now(),
            )
            created.append(stage_run)
            self._stage_runs.append(stage_run)
        return created

    def change_state(
        self, id: str, status: str, execution_id: Optional[str] = None
    ) -> bool:
        stage_run = self.get(id)
        stage_run.updated_at = datetime.now()

        if not stage_run.execution_id and execution_id is not None:
            stage_run.execution_id = execution_id

        if stage_run.status in end_status:
            return False

        if status in status_transitions.get(stage_run.status, []):
            stage_run.status = status
            return True

        return False

    def fork(
        self, stage_run: StageRun, custom_thread_data: Optional[str] = None
    ) -> StageRun:
        if stage_run.parent_id is None:
            return self.create_initial(stage_run.stage, {})
        parent_stage_run = self.get(stage_run.parent_id)
        new_stage_run = stage_run.clone_to_waiting()
        new_stage_run.data = parent_stage_run.data
        new_stage_run_dto = new_stage_run.to_dto()
        stage_runs = self.create_next(parent_stage_run, [new_stage_run_dto])
        if custom_thread_data:
            validate_json(custom_thread_data)
            stage_runs[0].data = json.loads(custom_thread_data)
        return stage_runs[0]

    def update_data(self, stage_run_id: str, data: dict) -> bool:
        stage_run = self.get(stage_run_id)
        stage_run.updated_at = datetime.now()
        stage_run.data = data
        return True


class RemoteStageRunRepository(StageRunRepository):
    def _request(
        self,
        method: str,
        path: str,
        body: Any = None,
        params: dict = {},
        raise_for_status: bool = True,
    ):
        headers: Mapping[str, str] = SIDECAR_HEADERS
        r = requests.request(
            method=method,
            url=f"{SIDECAR_URL}/stage-runs{path}",
            headers=headers,
            json=body,
            params=params,
        )

        if raise_for_status:
            r.raise_for_status()

        return r

    @staticmethod
    def create_from_dto(dto: dict) -> StageRun:
        return StageRun(
            id=dto["id"],
            stage=dto["stage"],
            data=dto["data"],
            status=dto["status"],
            parent_id=dto.get("parentId"),
            execution_id=dto.get("executionId"),
            created_at=from_utc_iso_string(dto["createdAt"]),
            updated_at=from_utc_iso_string(dto["updatedAt"]),
        )

    def get(self, id: str) -> StageRun:
        r = self._request("GET", path=f"/{id}")
        return self.create_from_dto(r.json())

    def find(self, filter: GetStageRunByQueryFilter) -> List[StageRun]:
        r = self._request("GET", path="/", params=filter.to_dict())
        return [self.create_from_dto(dto) for dto in r.json()]

    def find_ancestors(self, id: str) -> List[StageRun]:
        r = self._request("GET", path=f"/{id}/ancestors")
        return [self.create_from_dto(dto) for dto in r.json()]

    def create_initial(self, stage: str, data: dict = {}) -> StageRun:
        body = dict(data={}, stage=stage)
        r = self._request("POST", path="/", body=body)
        return self.create_from_dto(r.json())

    def create_next(self, parent: StageRun, dtos: List[Dict]) -> List[StageRun]:
        if len(dtos) == 0:
            return []

        parent_id = parent.id
        next_dtos = self.hydrate_next_dto(parent, dtos)
        r = self._request("PUT", path=f"/{parent_id}/children", body=next_dtos)
        return [self.create_from_dto(dto) for dto in r.json()]

    def change_state(
        self, id: str, status: str, execution_id: Optional[str] = None
    ) -> bool:
        r = self._request(
            "PATCH", path=f"/{id}", body={"status": status}, raise_for_status=False
        )

        if r.status_code == 409:
            return False

        r.raise_for_status()

        return True

    def find_leaves(
        self, filter: GetStageRunByQueryFilter, pagination: Pagination
    ) -> PaginatedListResponse:
        r = self._request(
            "GET",
            path="/leaves",
            params={**filter.to_dict(), **pagination.to_dict()},
            raise_for_status=False,
        )

        r.raise_for_status()

        return PaginatedListResponse.from_dict(r.json())

    def find_past(
        self, filter: GetStageRunByQueryFilter, pagination: Pagination
    ) -> PaginatedListResponse:
        raise NotImplementedError()

    def fork(self, stage_run: StageRun, custom_thread_data: Optional[str]) -> StageRun:
        raise NotImplementedError()

    def update_data(self, stage_run_id: str, data: dict) -> bool:
        r = self._request("PATCH", path=f"/{stage_run_id}", body={"data": data})
        if r.status_code == 409:
            return False

        r.raise_for_status()

        return True


def stage_run_repository_factory() -> StageRunRepository:
    if SIDECAR_URL is None:
        return LocalStageRunRepository()
    else:
        return RemoteStageRunRepository()



import abc
import datetime
import typing

import jwt
import requests

from ..utils.environment import PROJECT_ID, SIDECAR_HEADERS, SIDECAR_URL


def endcode_fake_jwt(email: str):
    return jwt.encode(
        key="fake",
        algorithm="HS256",
        payload={
            "email": email,
            "aud": PROJECT_ID,
            "exp": datetime.datetime.utcnow() + datetime.timedelta(days=7),
        },
    )


class AuthnRepository(abc.ABC):
    def authenticate(self, email: str) -> bool:
        raise NotImplementedError()

    def verify(self, email: str, token: str) -> typing.Optional[str]:
        raise NotImplementedError()


class ProductionAuthnRepository(AuthnRepository):
    base_url: str

    def __init__(self, sidecar_url: str):
        self.base_url = sidecar_url + "/passwordless"

    def authenticate(self, email: str) -> bool:
        url = self.base_url + "/authenticate"
        resp = requests.post(url, json={"email": email}, headers=SIDECAR_HEADERS)
        return resp.ok

    def verify(self, email: str, token: str) -> typing.Optional[str]:
        url = self.base_url + "/verify"
        resp = requests.post(
            url, json={"email": email, "token": token}, headers=SIDECAR_HEADERS
        )
        if not resp.ok:
            return None

        return resp.json()["jwt"]


class LocalAuthnRepository(AuthnRepository):
    emails = {}

    def authenticate(self, email: str) -> bool:
        self.emails[email] = True
        return True

    def verify(self, email: str, token: str) -> typing.Optional[str]:
        if token == "000000":
            # for testing purposes
            return None

        verified = self.emails.pop(email, False)
        if not verified:
            return None

        return endcode_fake_jwt(email)


def authn_repository_factory() -> AuthnRepository:
    if SIDECAR_URL is None:
        return LocalAuthnRepository()
    else:
        return ProductionAuthnRepository(SIDECAR_URL)



from typing import List, Optional, TypedDict
from ..file_utils import convert_file
from ..widget_base import Input, MultipleHandler


class CardOption(TypedDict):
    title: Optional[str]
    subtitle: Optional[str]
    image: Optional[str]
    description: Optional[str]
    topLeftExtra: Optional[str]
    topRightExtra: Optional[str]


class CardsInput(Input):
    type = 'cards-input'
    empty_value = None
    multiple: bool = False
    multiple_handler: MultipleHandler

    def __init__(self, key: str, label: str, options: List[CardOption], **
        kwargs):
        super().__init__(key)
        self.set_props(dict(label=label, options=options, **kwargs))

    def set_props(self, props):
        self.label = props.get('label', 'Label')
        self.options = [{**opt, 'image': convert_file(opt.get('image'))} for
            opt in props.get('options', [])]
        self.searchable = props.get('searchable', False)
        self.required = props.get('required', True)
        self.hint = props.get('hint', None)
        self.columns = props.get('columns', 2)
        self.full_width = props.get('full_width', False)
        self.layout = props.get('layout', 'list')
        self.disabled = props.get('disabled', False)
        self.multiple = props.get('multiple', False)
        self.empty_value = [] if self.multiple else None
        self.value = props.get('initial_value', self.empty_value)
        self.min = props.get('min', None)
        self.max = props.get('max', None)
        self.multiple_handler = MultipleHandler(self.multiple, self.min,
            self.max, self.required)

    def render(self, ctx: dict):
        return {'type': self.type, 'key': self.key, 'label': self.label,
            'hint': self.hint, 'options': self.options, 'multiple': self.
            multiple, 'searchable': self.searchable, 'required': self.
            required, 'columns': self.columns, 'fullWidth': self.full_width,
            'layout': self.layout, 'disabled': self.disabled, 'errors':
            self.errors, 'min': self.min, 'max': self.max, 'value': self.
            serialize_value()}

    def validate(self) ->List[str]:
        return self.multiple_handler.validate(self.value)

    def serialize_value(self) ->List:
        return self.multiple_handler.value_to_list(self.value)

    def parse_value(self, value):
        return self.multiple_handler.value_to_list_or_value(value)



from .PasswordInput import PasswordInput
from .DropdownInput import DropdownInput
from .CustomInput import CustomInput
from .ListInput import ListInput
from .TextareaInput import TextareaInput
from .RatingInput import RatingInput
from .NumberSliderInput import NumberSliderInput
from .CodeInput import CodeInput
from .PandasRowSelectionInput import PandasRowSelectionInput
from .PandasOutput import PandasOutput
from .FileOutput import FileOutput
from .CurrencyInput import CurrencyInput
from .RichTextInput import RichTextInput
from .TextOutput import TextOutput
from .ImageInput import ImageInput
from .NumberInput import NumberInput
from .ToggleInput import ToggleInput
from .VideoInput import VideoInput
from .LinkOutput import LinkOutput
from .CheckboxInput import CheckboxInput
from .PhoneInput import PhoneInput
from .NpsInput import NpsInput
from .EmailInput import EmailInput
from .ImageOutput import ImageOutput
from .MultipleChoiceInput import MultipleChoiceInput
from .HtmlOutput import HtmlOutput
from .CnpjInput import CnpjInput
from .MarkdownOutput import MarkdownOutput
from .ChecklistInput import ChecklistInput
from .TagInput import TagInput
from .ProgressOutput import ProgressOutput
from .IframeOutput import IframeOutput
from .CameraInput import CameraInput
from .LatexOutput import LatexOutput
from .CpfInput import CpfInput
from .TimeInput import TimeInput
from .CardsInput import CardsInput
from .TextInput import TextInput
from .FileInput import FileInput
from .PlotlyOutput import PlotlyOutput
from .DateInput import DateInput



from .debugpy import is_client_connected, start_debugger
from .launch_json import configure_launch_json, is_launch_json_configured
from .settings_json import configure_settings_json, is_settings_json_set

__all__ = [
    "configure_launch_json",
    "configure_settings_json",
    "is_client_connected",
    "is_launch_json_configured",
    "is_settings_json_set",
    "start_debugger",
]



import json
from typing import TYPE_CHECKING, Any, List, Union
from ..widget_base import Input, MultipleHandler
if TYPE_CHECKING:
    import pandas as pd


class PandasRowSelectionInput(Input):
    type = 'pandas-row-selection-input'
    empty_value: Union[List, Any] = None
    multiple: bool = False
    df: 'pd.DataFrame'

    def __init__(self, key: str, df: Any=None, **kwargs):
        super().__init__(key)
        self.set_props(dict(df=df, **kwargs))

    def set_props(self, props):
        self.df = props.get('df', '')
        self.required = props.get('required', True)
        self.hint = props.get('hint', None)
        self.full_width = props.get('full_width', False)
        self.display_index = props.get('display_index', False)
        self.disabled = props.get('disabled', False)
        self.label = props.get('label', '')
        self.filterable = props.get('filterable', False)
        self.multiple = props.get('multiple', False)
        self.empty_value = [] if self.multiple else None
        self.value = props.get('initial_value', self.empty_value)
        self.min = props.get('min', None)
        self.max = props.get('max', None)
        self.multiple_handler = MultipleHandler(self.multiple, self.min,
            self.max, self.required)

    def serialize_table(self):
        if self.df is None:
            import pandas as pd
            df = pd.DataFrame({'change the': [1, 2, 3], 'df property': [4, 
                5, 6]})
            df_json = df.to_json(orient='table')
            if not isinstance(df_json, str):
                raise Exception('df.to_json() did not return a string')
            serialized = json.loads(df_json)
            del serialized['schema']['pandas_version']
            return serialized
        df_json = str(self.df.to_json(orient='table'))
        serialized = json.loads(df_json)
        del serialized['schema']['pandas_version']
        return serialized

    def validate(self) ->List[str]:
        return self.multiple_handler.validate(self.value)

    def serialize_value(self) ->List:
        return self.multiple_handler.value_to_list(self.value)

    def parse_value(self, value):
        return self.multiple_handler.value_to_list_or_value(value)

    def render(self, ctx: dict):
        return {'type': self.type, 'key': self.key, 'hint': self.hint,
            'table': self.serialize_table(), 'required': self.required,
            'fullWidth': self.full_width, 'displayIndex': self.
            display_index, 'disabled': self.disabled, 'label': self.label,
            'multiple': self.multiple, 'filterable': self.filterable,
            'value': self.serialize_value(), 'errors': self.errors, 'min':
            self.min, 'max': self.max}



from abc import ABC, abstractmethod
from typing import Any, Dict, List, Literal, Optional

import requests
from pydantic.dataclasses import dataclass

from ..utils.environment import (
    SERVER_UUID,
    SIDECAR_HEADERS,
    SIDECAR_URL,
    WORKER_UUID,
)

ExecutionStatus = Literal["running", "lock-failed", "failed", "finished", "abandoned"]


@dataclass
class ExecutionDTO:
    id: str
    status: ExecutionStatus
    created_at: str
    context: Dict[str, Any]
    stage_id: str
    stage_run_id: Optional[str]


class ExecutionRepository(ABC):
    @abstractmethod
    def create(self, execution_dto: ExecutionDTO) -> None:
        raise NotImplementedError()

    @abstractmethod
    def update(self, execution_dto: ExecutionDTO) -> None:
        raise NotImplementedError()

    @abstractmethod
    def find(self, stage_run_id: str) -> List[ExecutionDTO]:
        raise NotImplementedError()


class LocalExecutionRepository(ExecutionRepository):
    def __init__(self):
        self.executions: Dict[str, ExecutionDTO] = {}

    def create(self, execution_dto: ExecutionDTO) -> None:
        self.executions[execution_dto.id] = execution_dto

    def update(self, execution_dto: ExecutionDTO) -> None:
        self.executions[execution_dto.id] = execution_dto

    def find(self, stage_run_id: str) -> List[ExecutionDTO]:
        return [
            execution
            for execution in self.executions.values()
            if execution.stage_run_id == stage_run_id
        ]


class RemoteExecutionRepository(ExecutionRepository):
    def __init__(
        self,
        url: str,
        headers: Dict[str, str],
    ):
        self.url = url
        self.headers = headers

    def create(self, execution_dto: ExecutionDTO) -> None:
        request_dto = dict(
            id=execution_dto.id,
            status=execution_dto.status,
            createdAt=execution_dto.created_at,
            context=execution_dto.context,
            stageId=execution_dto.stage_id,
            stageRunId=execution_dto.stage_run_id,
            workerId=WORKER_UUID(),
            appId=SERVER_UUID(),
        )

        res = requests.post(
            f"{self.url}/executions",
            json=request_dto,
            headers=self.headers,
        )

        res.raise_for_status()

    def update(self, execution_dto: ExecutionDTO) -> None:
        request_dto = dict(
            status=execution_dto.status,
            context=execution_dto.context,
            stageRunId=execution_dto.stage_run_id,
        )

        res = requests.patch(
            f"{self.url}/executions/{execution_dto.id}",
            json=request_dto,
            headers=self.headers,
        )

        res.raise_for_status()

    def find(self, stage_run_id: str) -> List[ExecutionDTO]:
        res = requests.get(
            f"{self.url}/executions",
            params=dict(stageRunId=stage_run_id),
            headers=self.headers,
        )

        res.raise_for_status()
        return [ExecutionDTO(**execution) for execution in res.json()]


def execution_repository_factory() -> ExecutionRepository:
    if SIDECAR_URL:
        return RemoteExecutionRepository(
            url=SIDECAR_URL,
            headers=SIDECAR_HEADERS,
        )
    else:
        return LocalExecutionRepository()



import abstra.workflows as aw

# Abstra Jobs are a simple way to schedule Python recurring tasks
# Setup the Schedule at the "Settings" tab

print("🕒 Job is running...")

# You can store data in the current thread,
# so it will be available in the next stages
aw.set_data("job_output", "Some data from the job!")



import os

HOST = os.getenv("ABSTRA_HOST", "localhost")
DEFAULT_PORT = os.getenv("PORT") or os.getenv("ABSTRA_SERVER_PORT")
ENABLE_TALISMAN = os.getenv("ABSTRA_ENABLE_TALISMAN", "false").lower() == "true"

BUILD_ID = os.getenv("ABSTRA_BUILD_ID") or "dev"
PROJECT_ID = os.getenv("ABSTRA_PROJECT_ID") or "dev-project-id"

IS_PRODUCTION = os.getenv("ABSTRA_ENVIRONMENT") == "production"
SHOW_WATERMARK = os.getenv("ABSTRA_SHOW_WATERMARK", "false") == "true"

CLOUD_API_ENDPOINT = os.getenv("CLOUD_API_ENDPOINT") or "https://cloud-api.abstra.cloud"
CLOUD_API_CLI_URL = f"{CLOUD_API_ENDPOINT}/cli"

PUBLIC_KEY = os.getenv("ABSTRA_JWT_PUBLIC_KEY_PEM")
FILES_FOLDER = os.getenv("ABSTRA_FILES_FOLDER")

# gunicorn.conf.py imports this variables, do not move them
SIDECAR_SHARED_TOKEN = os.getenv("ABSTRA_SIDECAR_SHARED_TOKEN", "shared")
SIDECAR_HEADERS = {"shared-token": SIDECAR_SHARED_TOKEN}
SIDECAR_URL = os.getenv("ABSTRA_SIDECAR_URL")


# gunicorn.conf.py imports this functions, do not move them
def WORKER_UUID():
    # Set during runtime
    return os.getenv("ABSTRA_WORKER_UUID")


def SERVER_UUID():
    # Set during runtime
    return os.getenv("ABSTRA_SERVER_UUID")



import logging
import os
import threading
import webbrowser

from ...server import get_local_app
from ...server.controller.main import MainController
from ...server.fs_watcher import watch_py_root_files
from ...settings import Settings
from ...stdio_monkey_patch import override_stdio
from ...utils.environment import HOST
from .messages import serve_message
from .version import check_latest_version


def editor(
    debug: bool,
    load_dotenv: bool,
):
    if not debug:
        log = logging.getLogger("werkzeug")
        log.setLevel(logging.WARNING)

    serve_message()
    check_latest_version()

    # used to block hackerforms lib from opening
    os.environ["ABSTRA_SERVER"] = "true"

    override_stdio(print_exceptions=False)

    controller = MainController()

    watch_py_root_files()

    port = Settings.server_port
    app = get_local_app(controller)
    threading.Timer(1, lambda: webbrowser.open(f"http://{HOST}:{port}/_editor")).start()
    app.run(
        host=HOST,
        port=port,
        debug=debug,
        load_dotenv=load_dotenv,
    )



from .server.controller.main import MainController
from .settings import SettingsController
from .stdio_monkey_patch import override_stdio


def cloud_setup(root: str):
    SettingsController.set_root_path(root)
    SettingsController.set_server_port(None)
    override_stdio(print_exceptions=True)

    return MainController()



from abc import ABC
from typing import TYPE_CHECKING, Any, TypedDict, Union


class LabelValueOption(TypedDict):
    label: str
    value: Any


AbstraOption = Union[str, LabelValueOption]

if TYPE_CHECKING:
    from pandas import DataFrame
    from PIL.Image import Image
    from plotly.graph_objects import Figure

    class PILImage(Image):
        pass

    class PandasDataFrame(DataFrame):
        pass

    PlotlyFigure = Figure
else:
    PandasDataFrame = ABC
    PILImage = ABC
    PlotlyFigure = ABC



from .execution import Execution
from .form_execution import FormExecution
from .hook_execution import HookExecution
from .job_execution import JobExecution

__all__ = ["FormExecution", "HookExecution", "JobExecution", "Execution"]



import os
import threading
import time
from datetime import datetime

from abstra_internals.logger import AbstraLogger

from ..contract.forms import FilesChangedMessage
from ..execution.form_execution import FormExecution
from ..modules import reload_project_local_modules
from ..repositories.project.project import Project, ProjectRepository


def has_local_dependencies_changed(project: Project, last_change: float) -> bool:
    for file in project.project_local_dependencies:
        if not file.exists():
            continue

        st_mtime = os.stat(file).st_mtime
        if st_mtime > last_change:
            return True
    return False


def reload_files_on_change(project: Project, last_change: float):
    if has_local_dependencies_changed(project, last_change):
        reload_project_local_modules()
        FormExecution.broadcast(FilesChangedMessage())
        return True
    else:
        return False


def files_changed_polling_loop():
    last_change = datetime.now().timestamp()

    while True:
        try:
            project = ProjectRepository.load()
            has_reloaded = reload_files_on_change(project, last_change)
            if has_reloaded:
                last_change = datetime.now().timestamp()
        except Exception as e:
            AbstraLogger.capture_exception(e)
            pass
        time.sleep(1)


def watch_py_root_files():
    threading.Thread(
        name="file_watcher", target=files_changed_polling_loop, daemon=True
    ).start()



from typing import Tuple

from .utils import quoted_identifier


class Comparator:
    def exp(self, column_name: str, idx: int) -> Tuple[str, list]:
        raise NotImplementedError()


class is_null(Comparator):
    def exp(self, column_name: str, idx: int) -> Tuple[str, list]:
        return f"{quoted_identifier(column_name)} IS NULL", []


class is_eq(Comparator):
    def __init__(self, value):
        self.value = value

    def exp(self, column_name: str, idx: int) -> Tuple[str, list]:
        return f"{quoted_identifier(column_name)}=${idx}", [self.value]


class is_neq(Comparator):
    def __init__(self, value):
        self.value = value

    def exp(self, column_name: str, idx: int) -> Tuple[str, list]:
        return f"{quoted_identifier(column_name)} <> ${idx}", [self.value]


class is_gt(Comparator):
    def __init__(self, value):
        self.value = value

    def exp(self, column_name: str, idx: int) -> Tuple[str, list]:
        return f"{quoted_identifier(column_name)} > ${idx}", [self.value]


class is_gte(Comparator):
    def __init__(self, value):
        self.value = value

    def exp(self, column_name: str, idx: int) -> Tuple[str, list]:
        return f"{quoted_identifier(column_name)} >= ${idx}", [self.value]


class is_lt(Comparator):
    def __init__(self, value):
        self.value = value

    def exp(self, column_name: str, idx: int) -> Tuple[str, list]:
        return f"{quoted_identifier(column_name)} < ${idx}", [self.value]


class is_lte(Comparator):
    def __init__(self, value):
        self.value = value

    def exp(self, column_name: str, idx: int) -> Tuple[str, list]:
        return f"{quoted_identifier(column_name)} <= ${idx}", [self.value]


class is_in(Comparator):
    def __init__(self, value):
        self.value = value

    def exp(self, column_name: str, idx: int) -> Tuple[str, list]:
        return f"{quoted_identifier(column_name)} IN ${idx}", [self.value]


class is_not_in(Comparator):
    def __init__(self, value):
        self.value = value

    def exp(self, column_name: str, idx: int) -> Tuple[str, list]:
        return f"{quoted_identifier(column_name)} NOT IN ${idx}", [self.value]


class is_like(Comparator):
    def __init__(self, value):
        self.value = value

    def exp(self, column_name: str, idx: int) -> Tuple[str, list]:
        return f"{quoted_identifier(column_name)} LIKE ${idx}", [self.value]


class is_not_like(Comparator):
    def __init__(self, value):
        self.value = value

    def exp(self, column_name: str, idx: int) -> Tuple[str, list]:
        return f"{quoted_identifier(column_name)} NOT LIKE ${idx}", [self.value]


class is_between(Comparator):
    def __init__(self, value1, value2):
        self.value1 = value1
        self.value2 = value2

    def exp(self, column_name: str, idx: int) -> Tuple[str, list]:
        return f"{quoted_identifier(column_name)} BETWEEN ${idx} AND ${idx+1}", [
            self.value1,
            self.value2,
        ]


class is_not_null(Comparator):
    def exp(self, column_name: str, idx: int) -> Tuple[str, list]:
        return f"{quoted_identifier(column_name)} IS NOT NULL", []


class is_ilike(Comparator):
    def __init__(self, value):
        self.value = value

    def exp(self, column_name: str, idx: int) -> Tuple[str, list]:
        return f"{quoted_identifier(column_name)} ILIKE ${idx}", [self.value]


class is_not_ilike(Comparator):
    def __init__(self, value):
        self.value = value

    def exp(self, column_name: str, idx: int) -> Tuple[str, list]:
        return f"{quoted_identifier(column_name)} NOT ILIKE ${idx}", [self.value]



import pathlib
import tempfile
import uuid
import zipfile

import requests

from ...cloud_api import create_build, update_build
from ...credentials import resolve_headers
from ...settings import Settings
from ...utils.file import files_from_directory


def _generate_zip_file() -> pathlib.Path:
    root_path = Settings.root_path
    zip_path = pathlib.Path(tempfile.gettempdir(), f"{uuid.uuid4()}.zip")

    with zipfile.ZipFile(zip_path, "w") as zip_file:
        for file in files_from_directory(root_path):
            zip_file.write(file, file.relative_to(root_path))

    return zip_path


def _upload_file(url: str, file_path: pathlib.Path):
    with file_path.open("rb") as f:
        requests.put(url=url, data=f.read())


def deploy():
    headers = resolve_headers()
    if not headers:
        print(
            "No project credentials found. Please run `abstra serve` and login via the web interface."
        )
        return

    data = create_build(headers)
    url = data.url
    build_id = data.build_id

    zip_path = _generate_zip_file()
    _upload_file(url=url, file_path=zip_path)
    update_build(headers=headers, build_id=build_id)
